<?php

namespace Service;

use Doctrine\DBAL\Connection;
use Doctrine\ORM\EntityManager;
use Exception\CustomException;
use Service\Integration\DEVOPSIntegration;
use Teknisa\Libs\Util\Repositories\Migration as MigrationRepository;
use Util\Utilities;
use Util\InstanceProvider;
use Factory\ProductServer as ProductServerFactory;
use Factory\Deploy as DeployFactory;
use Factory\Attachment as AttachmentFactory;
use Zeedhi\DevOps\Model\Configurationmanager\Entities\Deploy;
use Zeedhi\DevOps\Model\Configurationmanager\Entities\DeployStep;
use Zeedhi\DevOps\Model\Configurationmanager\Entities\DeployStepSrvDeploy;
use Zeedhi\DevOps\Model\Devops\Entities\ProductVersion;
use Zeedhi\DevOps\Model\Servermanager\Entities\DeployServer;
use Zeedhi\Framework\Remote\FileUploader\Server as ServerFileUploader;
use Zeedhi\DevOps\Model\Servermanager\Entities\ProductServer as ProductServerEntity;

class ProductDeploy
{
    const RECENT_VERSION = '*';
    const PRODUCT_KEY = 'product';

    /** @var EntityManager $entityManager */
    protected $entityManager;
    /** @var Utilities $utilities */
    protected $utilities;
    /** @var DEVOPSIntegration $devOpsIntegration */
    protected $devOpsIntegration;
    /** @var \Teknisa\Model\Teknisa\Repositories\OmProdutoOrg $omProdutoOrgRepository */
    protected $omProdutoOrgRepository;
    /** @var \Zeedhi\DevOps\Model\Servermanager\Repositories\Server $serverRepository */
    protected $serverRepository;
    /** @var \Zeedhi\DevOps\Model\Servermanager\Repositories\ProductServer $productServerRepository */
    protected $productServerRepository;
    /** @var \Zeedhi\DevOps\Model\Servermanager\Repositories\DeployServer $deployServerRepository */
    protected $deployServerRepository;
    /** @var \Zeedhi\DevOps\Model\Configurationmanager\Repositories\DeployPackage $deployPackageRepository */
    protected $deployPackageRepository;
    /** @var \Zeedhi\DevOps\Model\Devops\Repositories\ProductVersion $productVersionRepository */
    protected $productVersionRepository;
    /** @var \Zeedhi\DevOps\Model\Devops\Repositories\ProductVersionPackages $productVersionPackagesRepository */
    protected $productVersionPackagesRepository;
    /** @var \Zeedhi\DevOps\Model\Configurationmanager\Repositories\Deploy $deployRepository */
    protected $deployRepository;
    /** @var \Zeedhi\DevOps\Model\Configurationmanager\Repositories\Package $packageRepository */
    protected $packageRepository;
    /** @var \Zeedhi\DevOps\Model\Configurationmanager\Repositories\Product $productRepository */
    protected $productRepository;
    /** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\Resource $resourceRepository */
    protected $resourceRepository;
    /** @var \Zeedhi\DevOps\Model\Projectmanagement\Repositories\Attachment $attachmentRepository */
    private $attachmentRepository;
    /** @var ProductServerFactory $productServerFactory */
    protected $productServerFactory;
    /** @var AttachmentFactory $attachmentFactory */
    protected $attachmentFactory;
    /** @var DeployFactory $deployFactory */
    protected $deployFactory;
    /** @var ServerFileUploader $zhFileUploader */
    protected $zhFileUploader;

    protected $fileServerUrl;

    /**
     * @param EntityManager $entityManager
     * @param Utilities $utilities
     * @param DEVOPSIntegration $devOpsIntegration
     * @param ProductServerFactory $productServerFactory
     * @param DeployFactory $deployFactory
     * @param AttachmentFactory $attachmentFactory
     * @param ServerFileUploader $zhFileUploader
     * @param $fileServerUrl
     */
    public function __construct(EntityManager $entityManager, Utilities $utilities, DEVOPSIntegration $devOpsIntegration,
                                ProductServerFactory $productServerFactory, DeployFactory $deployFactory,
                                AttachmentFactory $attachmentFactory, ServerFileUploader $zhFileUploader, $fileServerUrl) {
        $this->utilities            = $utilities;
        $this->entityManager        = $entityManager;
        $this->devOpsIntegration    = $devOpsIntegration;
        $this->productServerFactory = $productServerFactory;
        $this->deployFactory        = $deployFactory;
        $this->fileServerUrl        = $fileServerUrl;
        $this->attachmentFactory    = $attachmentFactory;
        $this->zhFileUploader       = $zhFileUploader;

        $this->omProdutoOrgRepository   = $this->entityManager->getRepository('\Teknisa\Model\Teknisa\Entities\OmProdutoOrg');
        $this->serverRepository         = $this->entityManager->getRepository('\Zeedhi\DevOps\Model\Servermanager\Entities\Server');
        $this->deployServerRepository   = $this->entityManager->getRepository('\Zeedhi\DevOps\Model\Servermanager\Entities\DeployServer');
        $this->deployPackageRepository  = $this->entityManager->getRepository('\Zeedhi\DevOps\Model\Configurationmanager\Entities\DeployPackage');
        $this->productServerRepository  = $this->entityManager->getRepository('\Zeedhi\DevOps\Model\Servermanager\Entities\ProductServer');
        $this->productVersionRepository = $this->entityManager->getRepository('\Zeedhi\DevOps\Model\Devops\Entities\ProductVersion');
        $this->deployRepository         = $this->entityManager->getRepository('\Zeedhi\DevOps\Model\Configurationmanager\Entities\Deploy');
        $this->packageRepository        = $this->entityManager->getRepository('\Zeedhi\DevOps\Model\Configurationmanager\Entities\Package');
        $this->productRepository        = $this->entityManager->getRepository('\Zeedhi\DevOps\Model\Configurationmanager\Entities\Product');
        $this->resourceRepository       = $this->entityManager->getRepository('\Zeedhi\DevOps\Model\Projectmanagement\Entities\Resource');
        $this->attachmentRepository     = $this->entityManager->getRepository('\Zeedhi\DevOps\Model\Projectmanagement\Entities\Attachment');
        $this->productVersionPackagesRepository = $this->entityManager->getRepository('\Zeedhi\DevOps\Model\Devops\Entities\ProductVersionPackages');
    }

    /**
     * @param $conditions
     * @return array
     */
    public function findCustomerProducts($conditions, $productType,  $page, $pageSize) {
        $limit  = $page * $pageSize;
        $offset = ($page - 1) * $pageSize;
        $isTeknisa   = $this->utilities->currentClienteIsTeknisa();
        if($isTeknisa && empty($productType)){
            $productType = 'B';
        }else if(empty($productType)) {
            $productType = 'C';
        }
        $customerNrParc = $this->utilities->getCurrentCustomer()->getNrparcnegocio();
        $filterQuery = \Zeedhi\DevOps\Util\Utilities::createQueryFiltered($conditions);
        $cdcliente = $this->utilities->getCurrentCustomer()->getCdcliente();
        return $this->omProdutoOrgRepository->getOrgAllProductByNrOrgId($productType, $customerNrParc, $cdcliente, $filterQuery, $offset, $limit);
    }

    /**
     * @param $productId
     * @param $conditions
     * @return array
     */
    public function findCustomerServers($productId, $conditions) {
        $customerCode = $this->utilities->getCurrentCustomer()->getCdcliente();
        $filterQuery = \Zeedhi\DevOps\Util\Utilities::createQueryFiltered($conditions);
        return $this->serverRepository->findCustomerProductServers($productId, $customerCode, $filterQuery);
    }

    /**
     * @param $productServerId
     * @param $path
     * @param $packageId
     * @param $productId
     * @param $serverId
     * @param $backupLimit
     * @param $execScriptsAuto
     * @throws \Exception
     * @return int
     */
    public function saveCustomerServer($productServerId, $path, $packageId, $productId, $serverId, $backupLimit, $execScriptsAuto, $modulesIds) {
        $connection = $this->entityManager->getConnection();
        try {
            $this->entityManager->beginTransaction();
            /** @var \Zeedhi\DevOps\Model\Servermanager\Entities\ProductServer $productServer */
            $productServer = $productServerId ? $this->productServerRepository->find($productServerId) : null;
            /** @var \Zeedhi\DevOps\Model\Configurationmanager\Entities\Package $package */
            $package = $packageId ? $this->packageRepository->find($packageId) : null;
            /** @var \Zeedhi\DevOps\Model\Configurationmanager\Entities\Product $product */
            $product = $productId ? $this->productRepository->find($productId) : null;
            /** @var \Zeedhi\DevOps\Model\Servermanager\Entities\Server $server */
            $server = $serverId ? $this->serverRepository->find($serverId) : null;

            $productServer = $this->productServerFactory->save($productServer, $backupLimit, $path, $server, $product, $package, $execScriptsAuto);
            $this->entityManager->persist($productServer);
            $this->entityManager->flush();
            if($modulesIds){
                foreach($modulesIds as $modulesId){
                    /** @var \Zeedhi\DevOps\Model\Configurationmanager\Entities\Product $product */
                    $product = $modulesId ? $this->productRepository->find($modulesId) : null;
                    $moduleProductServer = $this->productServerRepository->findOneBy(array('product' => $product,
                                                                                        'usetype' => ProductServerEntity::TYPE_PRODUCTION,
                                                                                        'server' => $server));
                    $productModulesServer = $this->productServerFactory->save($moduleProductServer, $backupLimit, $path, $server, $product, $package, $execScriptsAuto);
                    $this->entityManager->persist($productModulesServer);
                }
            }
            $this->entityManager->flush();
            $connection->commit();

            return $productServer->getId();
        } catch(\Exception $e) {
            $connection->rollBack();
            throw $e;
        }
    }

    public function removeCustomerServer($productServerId) {
        $connection = $this->entityManager->getConnection();
        try {
            $this->entityManager->beginTransaction();
            /** @var \Zeedhi\DevOps\Model\Servermanager\Entities\ProductServer $productServer */
            $productServer = $this->productServerRepository->find($productServerId);

            $this->entityManager->remove($productServer);
            $this->entityManager->flush();
            $connection->commit();
        } catch(\Exception $e) {
            $connection->rollBack();
            throw $e;
        }
    }

    public function removeUnappliedDeploy($deployId) {
        $connection = $this->entityManager->getConnection();
        try {
            $this->entityManager->beginTransaction();
            $parentDeploy = $this->deployRepository->find($deployId);
            $deploys = $this->deployRepository->findBy(array('parent' => $deployId));
            array_push($deploys, $parentDeploy);

            foreach($deploys as $deploy){
                $deployPackage = $this->deployPackageRepository->findBy(array("deploy" => $deploy));
                foreach($deployPackage as $pkg) {
                    $this->entityManager->remove($pkg);
                }
                $deployServers = $this->deployServerRepository->findBy(array('deploy' => $deploy));
                foreach($deployServers as $server) {
                    $this->entityManager->remove($server);
                }
                $attachments = $this->attachmentRepository->findBy(array('deploy' => $deploy));
                foreach($attachments as $attachment) {
                    $this->entityManager->remove($attachment);
                }

                $this->entityManager->remove($deploy);
                $this->entityManager->flush();
            }
            $connection->commit();
        } catch(\Exception $e) {
            $connection->rollBack();
            throw $e;
        }
    }

    public function findDeployAudit($productId, $serverId, $startDate, $endDate, $isClient){
        $cdCliente  = $this->utilities->getCurrentCdCliente();
        $startDate .= ' 00:00:00';
        $endDate   .= ' 23:59:59';
        return $this->deployRepository->findDeployAudit($productId, $serverId, $startDate, $endDate, $isClient, $cdCliente);
    }


    public function findVersionAudit($productId, $moduleId,  $isClient, $initialVersion, $finalVersion, $initialRevision,$finalRevision,$initialBuild,$finalBuild){
         $cdCliente = $this->utilities->getCurrentCdCliente();
        return $this->deployRepository->findVersionAudit($productId, $moduleId, $isClient, $cdCliente,  $initialVersion, $finalVersion, $initialRevision,$finalRevision,$initialBuild,$finalBuild);
    }

    public function getLogToSend($labelFile){
        $fileContentString = '';
        $oldLogFiles = $this->attachmentRepository->findBy(array("label" => $labelFile));
        if (!empty($oldLogFiles)){
            foreach($oldLogFiles as $oldLogFile){
                $file = file_get_contents($this->fileServerUrl.$oldLogFile->getFileName());
                $temp = Utilities::getTemporaryDir();
                file_put_contents($temp.DIRECTORY_SEPARATOR.$labelFile, $file);
                $fileContentString .= $file;
            }
        } else {
            throw new CustomException('ERR_LOG_FAIL_TO_SEND');
        }
         return $fileContentString;
    }

    public function sendLogEmail($emails, $filename, $dir){
        if(!empty($emails)){
            foreach($emails as $email){
                InstanceProvider::getMailProductDeployMail()->sendattachMails($email, $filename, $dir);
            }
        }
    }

    private function getVersionModulesConfigurationKeys($versionModulesConfiguration) {
        $keys = array();
        $modulesVersion = @json_decode($versionModulesConfiguration, true);
        if($modulesVersion) {
            $keys = array_keys($modulesVersion);
        }
        return $keys;
    }

    /**
     * @param $modulesVersion
     * @param $productId
     * @param $productVersionId
     * @param $previousDeploy
     * @throws \Exception
     * @return array
     */
    public function getAvailableVersionInfo($modulesVersion, $productId, $productVersionId, $previousDeploy) {

        $organizationId  = InstanceProvider::getEnvironment()->getCurrentOrganizationId();
        $productNotFound = !$modulesVersion ? true : false;
        /** @var \Zeedhi\DevOps\Model\Configurationmanager\Entities\Product $product */
        $product = $this->productRepository->find($productId);
        /** @var ProductVersion $productVersion */
        $productVersion = $this->productVersionRepository->find($productVersionId);
        $currentBuild   = $this->productVersionRepository->findCurrentBuildFromVersion($productId, $organizationId,
            $productVersion->getVersion(), $productVersion->getRevision());
        $productVersion = $this->productVersionRepository->find($currentBuild[0]['ID']);
        $modulesConfiguration = $productVersion->getModulesConfiguration() ?: $product->getModulesConfiguration();
        $versionKeys = $this->getVersionModulesConfigurationKeys($productVersion->getModulesConfiguration());

        if($modulesConfiguration) {
            $productModulesVersion = json_decode($modulesConfiguration, true);
            if($modulesVersion) {
                foreach($productModulesVersion as $key => $productModule) {
                    $keyExists = false;
                    foreach($modulesVersion as &$module) {
                        $moduleDevOpsId = isset($module['DEVOPS_ID']) ? $module['DEVOPS_ID'] : $module['devOpsId'];
                        $moduleKey      = isset($module['KEY']) ? $module['KEY'] : null;

                        if($moduleDevOpsId == $productModule['devOpsId']) {
                            if(isset($productModule['maxVersion'])) {
                                $module['MAX_VERSION'] = $productModule['maxVersion'];
                            }
                            if(isset($productModule['nextProduct'])) {
                                $module['nextProduct'] = $productModule['nextProduct'];
                            }
                            if(isset($productModule['nextModule'])) {
                                $module['nextModule'] = $productModule['nextModule'];
                            }
                        }

                        if(empty($moduleKey) || $moduleKey == $key) {
                            $keyExists = true;
                        }
                    }
                    if(!$keyExists && (!isset($productModule['onlyControl']) || !$productModule['onlyControl'])) {
                        $modulesVersion[] = array(
                            'KEY'         => $key,
                            'MAX_VERSION' => isset($productModule['maxVersion']) ? $productModule['maxVersion'] : null,
                            'DEVOPS_ID'   => $productModule['devOpsId'],
                            'nextProduct' => isset($productModule['nextProduct']) ? $productModule['nextProduct'] : null,
                            'nextModule'  => isset($productModule['nextModule']) ? $productModule['nextModule'] : null,
                        );
                    }
                }
            } else {
                $modulesVersion = $productModulesVersion;
            }
        }
        if(!$modulesVersion) {
            throw new CustomException('ERR_PRODUCT_MODULES_NOT_FOUND');
        }
        $availableOrgDevOpsId = $this->getAvailableOrgDevOpsId($productId);
        foreach($modulesVersion as $key => &$module) {
            $devOpsId    = isset($module['DEVOPS_ID'])   ? $module['DEVOPS_ID']   : $module['devOpsId'];
            $onlyControl = isset($module['onlyControl']) ? $module['onlyControl'] : false;
            $nextProduct = isset($module['nextProduct']) ? $module['nextProduct'] : false;
            $nextModule  = isset($module['nextModule'])  ? $module['nextModule']  : false;
            $isTeknisa   = $this->utilities->currentClienteIsTeknisa();
            if($onlyControl ||
               (!in_array($devOpsId, $availableOrgDevOpsId) && !empty($availableOrgDevOpsId) && !$isTeknisa && !$nextProduct && !$nextModule) ||
               (isset($module['KEY']) && !empty($versionKeys) && !in_array($module['KEY'], $versionKeys)) ||
               (!empty($previousDeploy) && !isset($previousDeploy[$devOpsId]))
            ) {
                unset($modulesVersion[$key]);
            } else {
                if((isset($module['KEY']) && $module['KEY'] == self::PRODUCT_KEY) || $key == self::PRODUCT_KEY) {
                    $module['MAX_VERSION'] = $productVersion->getVersion() . '.' . $productVersion->getRevision();
                }
                if(isset($module['MAX_VERSION']) || isset($module['maxVersion'])) {
                    $modulesMaxVersion = isset($module['MAX_VERSION']) ? $module['MAX_VERSION'] : $module['maxVersion'];
                } else {
                    $modulesMaxVersion = null;
                }

                $maxVersion = $this->getMaxVersion($modulesMaxVersion, $devOpsId, $previousDeploy);
                $this->factoryModuleProperties($module, $devOpsId, $modulesMaxVersion, $key, $maxVersion, $productNotFound);
            }
        }
        return array_values($modulesVersion);
    }

    private function getAvailableOrgDevOpsId($devOpsId) {
        $cdCliente = InstanceProvider::getUtilities()->getCurrentCdCliente();
        $elements  = InstanceProvider::getRepository()->getOmProdutoElemento()->getProductElementsByProductIdAndCdcliente($devOpsId, $cdCliente);

        $availableOrgDevOpsId = array();
        foreach ($elements as $element) {
            if(!in_array($element['NREXTERNAL_PRODUCT_ID'], $availableOrgDevOpsId)) {
                array_push($availableOrgDevOpsId, $element['NREXTERNAL_PRODUCT_ID']);
            }
        }

        return $availableOrgDevOpsId;
    }


    function getProductModules($productId) {
        $customerCode = InstanceProvider::getUtilities()->getCurrentCdCliente();


        $productModules = $this->deployRepository->getProductModules($productId, $customerCode);
        return $productModules;
    }

    /**
     * @param $productId
     * @param $filter
     * @return array
     */
    public function findCustomerDeploys($productId, $filter) {
        $customerCode = $this->utilities->getCurrentCdCliente();
        $deployDate   = Utilities::getValueInFilterDataSourceByName($filter, 'DEPLOY_DATE', true);
        $id           = Utilities::getValueInFilterDataSourceByName($filter, 'ID', false);
        $filterQuery  = \Zeedhi\DevOps\Util\Utilities::createQueryFiltered($filter);
        if($deployDate) {
            $init = $deployDate[0] . ' 00:00:00';
            $end  = $deployDate[1] . ' 23:59:59';
            $filterQuery .= " AND (DEPLOY_DATE BETWEEN TO_DATE('$init', 'DD/MM/YYYY HH24:MI:SS')
                                                   AND TO_DATE('$end',  'DD/MM/YYYY HH24:MI:SS')
                                   OR DEPLOY_DATE IS NULL)";
        }

        if(empty($productId) && $id) {
            $deployEntity = $this->deployRepository->find($id);
            $productId = $deployEntity->getProduct()->getId();
        }
        $customerDeploys = $this->deployRepository->findCustomerDeploys($productId, $customerCode, $filterQuery);

        foreach ($customerDeploys as &$deploy) {
            /** @var \Zeedhi\DevOps\Model\Configurationmanager\Entities\DeployPackage[] $deployPackages */
            $deployPackages = $this->deployPackageRepository->findBy(array('deploy' => $deploy['ID']));
            foreach($deployPackages as $deployPackage) {
                $prodVersionPackage = $deployPackage->getProductVersionPackage();
                $deploy['PACKAGE'][] = array(
                    'ID' => $prodVersionPackage->getId(),
                    'URL' => $prodVersionPackage->getUrl(),
                    'NAME' => $prodVersionPackage->getName(),
                    'FILE_NAME' => $prodVersionPackage->getFileName(),
                    'PATH_TO_DEPLOY' => $prodVersionPackage->getPathToDeploy()
                );
            }

            if($deploy['PRODUCT_SERVER_ID']) {
                $maxDateByServers = $this->deployRepository->getMaxDeployDateByServers($productId, $deploy['PRODUCT_SERVER_ID']);
                $deploy['MAX_DEPLOY_DATE'] = $maxDateByServers[0]["MAX_DATE"];
            }
            $this->setValueDowngrade($deploy);
            if (!empty($deploy['TICKET_ID']) && !empty($deploy['CD_SYSTEM_TICKET']) && $deploy['CD_SYSTEM_TICKET'] == Deploy::CD_SIS_TICKET_MOVIDESK) {
                $ticket = InstanceProvider::getService()->getMovideskIntegration()->findTicketById($deploy['TICKET_ID']);
                if (!empty($ticket)) {
                    $deploy['TICKET_NAME'] = $ticket["subject"];
                }
            } else {
                $deploy['TICKET_ID'] = null;
            }
        }
        return $customerDeploys;
    }

    /**
     * @param $deployStepId
     * @return array
     */
    public function findCustomerDeploySteps($deployStepId)
    {
        $customerDeploysSteps = $this->deployRepository->findCustomerDeploySteps($deployStepId);
        foreach ($customerDeploysSteps as &$deploy) {
            $this->setValueDowngrade($deploy);
        }
        return  $customerDeploysSteps;
    }

    public function setValueDowngrade(&$row)
    {
        if ($row['VERSION'] !== '-' && $row['PREVIOUS_VERSION'] !== '-') {
            $v = explode('.', $row['VERSION']);
            $pv = explode('.', $row['PREVIOUS_VERSION']);
            $version = intval(str_pad($v[0], 5, '0', STR_PAD_LEFT) . str_pad($v[1], 5, '0', STR_PAD_LEFT) . str_pad($v[2], 5, '0', STR_PAD_LEFT));
            $availableVersion = intval(str_pad($pv[0], 5, '0', STR_PAD_LEFT) . str_pad($pv[1], 5, '0', STR_PAD_LEFT) . str_pad($pv[2], 5, '0', STR_PAD_LEFT));
            $row['DOWNGRADE'] = $availableVersion > $version ? 'Y' : 'N';
        } else {
            $row['DOWNGRADE'] = 'N';
        }
    }

    /**
     * @param $deployId
     * @param $filter
     * @return array
     */
    public function findDeployMigrations($deployId, $filter) {
        $filterQuery = \Zeedhi\DevOps\Util\Utilities::createQueryFiltered($filter);
        return $this->attachmentRepository->getDeployMigrationsScripts($deployId, $filterQuery);
    }

    /**
     * @param $productId
     * @param $filter
     * @return array
     */
    public function findProductPackages($productId, $filter) {
        $filterQuery = \Zeedhi\DevOps\Util\Utilities::createQueryFiltered($filter);
        return $this->packageRepository->findProductPackages($productId, $filterQuery);
    }

     /**
     * @param $productId
     * @param $productServerId
     * @param $productModules
     * @param $productPath
     * @param $checkedModulesIds
     * @param $deployStepServerId
     * @param $mainDeployId
     * @throws \Exception
     * @return array
     */
    public function getApplyDeployInfo($productId, $productServerId, $productModules, $productPath, $checkedModulesIds, $deployStepServerId = null, $mainDeployId = null) {
        $connection = $this->entityManager->getConnection();
        $this->entityManager->beginTransaction();
        try {
            $productIds = $deployIds = $deploy = array();
            $customerCode = $this->utilities->getCurrentCdCliente();
            if($deployStepServerId) {
                /** @var \Zeedhi\DevOps\Model\Configurationmanager\Entities\DeployStepServer $deployStepServer */
                $deployStepServer = InstanceProvider::getRepository()->getDeployStepServer()->find($deployStepServerId);
                if($deployStepServer->getDeploy() && $deployStepServer->getDeploy()->getProduct()->getId() == $productId) {
                    $deploy = array (array (
                        'ID'   => $deployStepServer->getDeploy()->getId(),
                        'NAME' => $deployStepServer->getDeploy()->getName()
                    ));
                }
            } else {
                $deploy = $this->deployRepository->findOpenedCustomerDeploy($productId, $customerCode, $productServerId);
            }
            /** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\Resource $customerResource */
            $customerResource = $this->resourceRepository->findOneBy(array('externalCode' => $customerCode));

            $product = array_filter($productModules, function($module) {
                return empty($module['PARENT_ID']);
            })[0];

            if(empty($deploy)) {
                $deployResponse = $this->devOpsIntegration->createDeploy($productId, $productServerId, $product,
                    $productPath, true);
                $deployId = $deployResponse['savedDeploy']['ID'];
                /** @var Deploy $productDeploy */
                $productDeploy = $this->deployRepository->find($deployId);
                $deployName = $productDeploy->getName();
                $productDeploy->setIsCustomerDeploy(Deploy::IS_CUSTOMER_DEPLOY);
                $productDeploy->setResource($customerResource);
                if($deployStepServerId) {
                    $productDeploy->setIsDeployStep(Deploy::IS_DEPLOY_STEP);
                }
            } else {
                $deployId   = $deploy[0]['ID'];
                $deployName = $deploy[0]['NAME'];
                /** @var Deploy $productDeploy */
                $productDeploy = $this->deployRepository->find($deployId);
            }
            /** @var ProductVersion $productVersion */
            $productVersion = $this->productVersionRepository->findOneBy(array('id' => $product['ACTUAL_PRODUCT_VERSION_ID']));
            $productDeploy->setPreviousProductVersion($productVersion);
            if($mainDeployId) {
                $productDeploy->setMainDeployId($mainDeployId);
            }
            $this->entityManager->persist($productDeploy);

            if($deployStepServerId) {
                $this->setDeployInDeployStepServer($deployStepServerId, $deployId);
            }

            foreach($productModules as &$module) {
                /** @var \Zeedhi\DevOps\Model\Devops\Entities\ProductVersionPackages[] $prodVersionPackages */
                $prodVersionPackages = $this->productVersionPackagesRepository->findBy(array('productVersion' => $module['PRODUCT_VERSION_ID']));
                if($module['ID'] == $product['ID']) {
                    $productPackages = array();
                    foreach($prodVersionPackages as $package) {
                        if($package->getPathToDeploy() == $productPath) {
                            $productPackages[] = $package;
                        }
                    }
                    $prodVersionPackages = $productPackages;
                }

                foreach($prodVersionPackages as $prodVersionPackage) {
                    $module['PACKAGE'][] = array(
                        'ID' => $prodVersionPackage->getId(),
                        'URL' => $prodVersionPackage->getUrl(),
                        'NAME' => $prodVersionPackage->getName(),
                        'FILE_NAME' => $prodVersionPackage->getFileName(),
                        'PATH_TO_DEPLOY' => $prodVersionPackage->getPathToDeploy()
                    );
                }

                $productIds[] = $module['ID'];
                if(empty($module['PARENT_ID'])) {
                    $module['DEPLOY_ID']   = $deployId;
                    $module['DEPLOY_NAME'] = $deployName;
                    $deployIds[] = $deployId;
                } else {
                    $deployModule = $this->deployRepository->findOneBy(array('parent' => $deployId, 'product' => $module['ID']));
                    if(empty($deployModule)) {
                        $deployModule = $this->createModuleDeploy($module, $deployId, $customerResource, $deployStepServerId);
                    }
                    $module['DEPLOY_ID']   = $deployModule->getId();
                    $module['DEPLOY_NAME'] = $deployModule->getName();
                    $deployIds[] = $deployModule->getId();
                }
            }
            $this->entityManager->flush();
            $connection->commit();

            $checkedModulesIds = empty($checkedModulesIds) ? $productIds : $checkedModulesIds;
            return array (
                'PRODUCT_MODULES'   => $productModules,
                'REPOSITORIES_INFO' => InstanceProvider::getRepository()->getOmProdutoRepositorio()->getMigrationRepositoriesInfoWithModules($checkedModulesIds, $deployIds),
                'CURRENT_USER_DATA' => array (
                    'CURRENT_ORG_ID'        => $this->getOrganizationIdByCurrentCdCliente(),
                    'CURRENT_OPERATOR_CODE' => InstanceProvider::getEnvironment()->getCurrentUserId()
                ),
                'MIGRATION_QUERIES' => array(
                    'INSERT' => MigrationRepository::INSERT_LAST_MIG_QUERY,
                    'UPDATE' => MigrationRepository::UPDATE_LAST_MIG_QUERY,
                    'PENDENT_FILE' => MigrationRepository::PENDENT_FILE_MIG_QUERY
                )
            );
        } catch(\Exception $e) {
            if($connection->isTransactionActive()) {
                $connection->rollBack();
            }
            throw $e;
        }
    }

    public function validateCreateDeploy($productServerId) {
        $deploySteps = InstanceProvider::getRepository()->getDeployStep()->findPendingDeployStepsByProductServer($productServerId);
        if(!empty($deploySteps)) {
            throw new CustomException('ALERT_EXISTS_DEPLOY_STEP_WITH_SERVER');
        }
        return true;
    }

    public function validateDeployPermition($productId) {
        $idPerDeploy = InstanceProvider::getRepository()->getDeployStep()->validateDeployPermition($productId);
        $isTeknisa   = $this->utilities->currentClienteIsTeknisa();
        if($idPerDeploy[0]["IDPERDEPLOY"] == 'N' && !$isTeknisa) {
            throw new CustomException('ALERT_PRODUCT_UNALLOWED_FOR_DEPLOY');
        }
        return true;
    }

    public function getOrganizationIdByCurrentCdCliente() {
        $cdCliente = InstanceProvider::getUtilities()->getCurrentCdCliente();
        /** @var \Teknisa\Model\Geral\Entities\Cliente $cliente */
        $cliente = InstanceProvider::getRepository()->getCliente()->findOneBy(array('cdcliente' => $cdCliente));
        /** @var \Teknisa\Model\Geral\Entities\Organizacao $organizacao */
        $organizacao = InstanceProvider::getRepository()->getOrganizacao()->findOneBy(array('nrparcnegocio' => $cliente->getNrparcnegocio()));

        return $organizacao->getNrorg();
    }

    /**
     * @param $productId
     * @param $productServerId
     * @param $productVersionId
     * @param $serverIsWindows
     * @return mixed
     * @throws \Exception
     */
    public function getProductDeployScripts($productId, $productServerId, $productVersionId, $serverIsWindows) {
        $scriptResponse = $this->devOpsIntegration->getProductDeployScripts($productId, $productServerId, $productVersionId, $serverIsWindows);
        return $scriptResponse['product_files'];
    }

    /**
     * @param $productId
     * @param $versionCheck
     * @param $urlWebService
     * @throws \Exception
     * @return int
     */
    public function saveCustomerProduct($productId, $versionCheck, $urlWebService) {
        $connection = $this->entityManager->getConnection();
        try {
            $this->entityManager->beginTransaction();

            /** @var \Zeedhi\DevOps\Model\Configurationmanager\Entities\Product $product */
            $product = $this->productRepository->find($productId);
            $product->setVersionCheckType($versionCheck);
            $product->setUrlWebservice($urlWebService);

            $this->entityManager->persist($product);
            $this->entityManager->flush();
            $connection->commit();

            return $product->getId();
        } catch(\Exception $e) {
            $connection->rollBack();
            throw $e;
        }
    }

    /**
     * @param $module
     * @param $devOpsId
     * @param $modulesMaxVersion
     * @param $key
     * @param $maxVersion
     * @param $productNotFound
     */
    private function factoryModuleProperties(&$module, $devOpsId, $modulesMaxVersion, $key, $maxVersion, $productNotFound) {
        $version = '-';
        if(isset($module['VERSION'])) {
            $version = $module['VERSION'];
        } else if(isset($module['version'])) {
            $version = $module['version'];
        }
        $product = $this->productRepository->find($devOpsId);
        $module['DEVOPS_ID']   = $devOpsId;
        $module['MAX_VERSION'] = $modulesMaxVersion;
        $module['VERSION']     = $productNotFound ? '-' : $version;
        $module['KEY']         = isset($module['KEY']) ? $module['KEY'] : $key;
        $module['NAME']        = $product ? $product->getName() : '-';
        $module['AVAILABLE_VERSION']  = empty($maxVersion) ? '-' : $maxVersion[0]['VERSION'];
        $module['PRODUCT_VERSION_ID'] = empty($maxVersion) ? null : $maxVersion[0]['PRODUCT_VERSION_ID'];
        $module['ACTUAL_PRODUCT_VERSION_ID'] = $productNotFound ? null : $this->findActualProductVersionId($devOpsId, $version);
    }

    /**
     * @param $productId
     * @param $version
     * @return int|null
     */
    private function findActualProductVersionId($productId, $version)
    {
        if($version === '-') {
            return null;
        }
        $fullVersion = explode('.', $version);

        $productVersion = $this->productVersionRepository->findOneBy(
            array('product' => $productId, 'version' => $fullVersion[0], 'revision' => $fullVersion[1], 'build' => $fullVersion[2])
        );
        return $productVersion ? $productVersion->getId() : null;
    }

    /**
     * @param $modulesMaxVersion
     * @param $devOpsId
     * @param $previousDeploy
     * @return array
     */
    private function getMaxVersion($modulesMaxVersion, $devOpsId, $previousDeploy) {
        $versionFilter = " AND ((PR.IS_LTS = 'Y' AND PR.IS_REVISION = 'Y') OR 
                                (PR.IS_LTS = 'N' AND PR.IS_REVISION = 'N'))";
        if(!empty($previousDeploy)) {
            $versionFilter = ' AND PR.ID = ' . $previousDeploy[$devOpsId];
        } else if (!empty($modulesMaxVersion)) {
            $actualVersion = explode('.', $modulesMaxVersion);
            foreach ($actualVersion as $index => $version) {
                $versionColumn = 'VERSION';
                if ($index == 1) {
                    $versionColumn = 'REVISION';
                } else if ($index == 2) {
                    $versionColumn = 'BUILD';
                }
                if (trim($version) == self::RECENT_VERSION) {
                    break;
                } else {
                    $versionFilter .= ' AND ' . $versionColumn . ' = ' . $version;
                }
            }
        }
        return $this->productVersionRepository->findMaxProductVersionWithFilter($devOpsId, $versionFilter);
    }

    /**
     * @param $module
     * @param $deployId
     * @param $customerResource
     * @param $deployStepServerId
     * @return Deploy
     */
    private function createModuleDeploy($module, $deployId, $customerResource, $deployStepServerId)
    {
        /** @var \Zeedhi\DevOps\Model\Configurationmanager\Entities\Product $product */
        $product = $this->productRepository->find($module['ID']);
        /** @var ProductVersion $productVersion */
        $productVersion = $this->productVersionRepository->find($module['PRODUCT_VERSION_ID']);
        /** @var ProductVersion $productVersion */
        $previousProductVersion = $this->productVersionRepository->findOneBy(array('id' => $module['ACTUAL_PRODUCT_VERSION_ID']));
        /** @var Deploy $parentDeploy */
        $parentDeploy = $this->deployRepository->find($deployId);

        $deployModule = $this->deployFactory->save(null, $product, $productVersion, $customerResource, $parentDeploy,
            $module['NAME'], Deploy::IS_ACTIVE_YES, null, Deploy::PENDING_DEPLOY_STATUS, null, Deploy::IS_CUSTOMER_DEPLOY, $previousProductVersion);
        if($deployStepServerId) {
            $deployModule->setIsDeployStep(Deploy::IS_DEPLOY_STEP);
        }
        $this->entityManager->persist($deployModule);

        return $deployModule;
    }

    /**
     * @param $productServerId
     * @param $currentDeployId
     * @param $removedPaths
     * @param $backupLimit
     * @param $productId
     * @throws \Doctrine\DBAL\ConnectionException
     * @throws \Exception
     */
    public function inactiveDeploysRemoved($productServerId, $removedPaths, $currentDeployId, $backupLimit, $productId) {
        $connection = $this->entityManager->getConnection();
        $connection->beginTransaction();
        try {
            if($backupLimit == 0) {
                $oldServers = $this->deployServerRepository->findOldDeployServerToInactive($productId, $productServerId, $currentDeployId);
                foreach($oldServers as $oldServer) {
                    /** @var DeployServer $deployServer */
                    $deployServer = $this->deployServerRepository->find($oldServer['DEPLOY_SERVER_ID']);
                    $deployServer->setIsActive(Deploy::IS_ACTIVE_NO);
                    $this->entityManager->persist($deployServer);
                    $this->inactiveDeploy($oldServer['DEPLOY_ID'], $deployServer);
                }
            }

            foreach($removedPaths as $path) {
                $path = explode('DEP_', $path);
                $path = explode('_', $path[1]);
                $deployId = $path[0];
                if($deployId != $currentDeployId) {
                    /** @var DeployServer $deployServer */
                    $deployServer = $this->deployServerRepository->findOneBy(array('deploy' => $deployId,
                        'productServer' => $productServerId));
                    $deployServer->setIsActive(Deploy::IS_ACTIVE_NO);
                    $this->entityManager->persist($deployServer);
                    $this->inactiveDeploy($deployId, $deployServer);
                }
            }

            $this->entityManager->flush();
            $connection->commit();
        } catch (\Exception $e) {
            $connection->rollBack();
            throw $e;
        }
    }

    /**
     * @param $deployId
     * @param DeployServer $deployServer
     */
    private function inactiveDeploy($deployId, DeployServer $deployServer)
    {
        $allDeployServers = $this->deployServerRepository->findBy(array('deploy' => $deployId));
        $isInactiveDeploy = true;
        /** @var DeployServer $aDeployServer */
        foreach ($allDeployServers as $aDeployServer) {
            if ($aDeployServer->getIsActive() == Deploy::IS_ACTIVE_YES &&
                $aDeployServer->getId() != $deployServer->getId()
            ) {
                $isInactiveDeploy = false;
            }
        }

        if ($isInactiveDeploy) {
            /** @var Deploy $deploy */
            $deploy = $this->deployRepository->find($deployId);
            $deploy->setIsActive(Deploy::IS_ACTIVE_NO);
            $this->entityManager->persist($deploy);

            $childrenDeploy = $this->deployRepository->findBy(array('parent' => $deploy));
            /** @var Deploy $children */
            foreach($childrenDeploy as $children) {
                $children->setIsActive(Deploy::IS_ACTIVE_NO);
                $this->entityManager->persist($deploy);
            }
        }
    }

    /**
     * @param array $documents
     * @param $deployId
     * @param null $type
     * @return array
     * @throws \Exception
     */
    public function saveLog(array $documents, $deployId, $type = null){
        try {
            $deploy = $this->deployRepository->find($deployId);
            return $this->updateLog($documents, $deploy, $type);
        } catch (\Exception $e){
            throw $e;
        }
    }

    /**
     * @param $labelFile
     * @return string
     */
    public function getFullLog($labelFile){
        $fileContentString = '';
        /** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\Attachment[] $oldLogFiles */
        $oldLogFiles = $this->attachmentRepository->findBy(array("label" => $labelFile));
        foreach($oldLogFiles as $oldLogFile){
            $file = file_get_contents($this->fileServerUrl . $oldLogFile->getFileName());
            $fileContentString .= $file;
        }
        return $fileContentString;
    }

    public function downloadDeployMigration($id) {
        /** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\Attachment $attachment */
        $attachment = $this->attachmentRepository->findOneBy(array("id" => $id));
        $fileDir  = Utilities::getTemporaryDir();
        $fileName = $attachment->getLabel();
        $script = mb_convert_encoding( $attachment->getDeployScript(), "ISO-8859-1");
        file_put_contents($fileDir . DIRECTORY_SEPARATOR . $fileName, $script);
        Utilities::downloadAndRemoveFile($fileDir . DIRECTORY_SEPARATOR . $fileName, $fileName, 'text/plain', true);
    }

    /**
     * @param array $documents
     * @param $deploy
     * @param $type
     * @return array
     * @throws \Exception
     */
    public function updateLog(array $documents, $deploy, $type){
        try {
            $filesUploaded = array();
            foreach ($documents as $fileData) {
                $labelFile = empty($fileData['name']) ? $fileData['NAME'] : $fileData['name'];
                /** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\Attachment $oldLogFile */
                $oldLogFile = $this->attachmentRepository->findOneBy(array("label" => $labelFile, "deploy" => $deploy));

                $typeFile = empty($fileData['type']) ? $type : $fileData['type'];
                $sizeFile = empty($fileData['size']) ? null : $fileData['size'];
                $pathFile = empty($fileData['path']) ? null : $fileData['path'];
                $deployScript = empty($fileData['SCRIPT']) ? null : $fileData['SCRIPT'];

                if (is_null($oldLogFile)) {
                    $attachment = $this->attachmentFactory->createAttachment($typeFile, $pathFile, $labelFile, $sizeFile, $deploy, $deployScript);
                } else {
                    if($pathFile){
                        $this->zhFileUploader->deleteFile($oldLogFile->getFileName());
                    }
                    $attachment = $this->attachmentFactory->updateAttachmentPath($oldLogFile, $pathFile, $deployScript);
                }
                $this->entityManager->persist($attachment);
                $filesUploaded[] = $attachment->getId();
            }
            $this->entityManager->flush();
            return $filesUploaded;
        } catch (\Exception $e){
            throw $e;
        }
    }

    public function updatesSetCdSystemTicket($deployId, Connection $connection, $ticket)
    {
        if($ticket){
            $sql = <<<SQL
        UPDATE DEPLOY SET CD_SYSTEM_TICKET = :CDSYSTEM WHERE ID = :DEPLOY_ID
SQL;
            $params = array(
                ":CDSYSTEM" => Deploy::CD_SIS_TICKET_MOVIDESK,
                ":DEPLOY_ID" => $deployId
            );
            $std = $connection->prepare($sql);
            $std->execute($params);
        }
    }

    /**
     * @param $deployId
     * @param $date
     * @param $isActive
     * @param $pathBkp
     * @param $status
     * @param $productModules
     * @param $ticket
     * @param $server
     * @param $version
     * @throws \Exception
     * @return array
     */
    public function saveDeploy($deployId, $date, $isActive, $pathBkp, $status, $productModules, $ticket, $server, $version) {
        $connection = $this->entityManager->getConnection();
        $connection->beginTransaction();
        try {
            /** @var Deploy $deploy */
            $deploy = $this->deployRepository->find($deployId);

            if($date) {
                $date = \DateTime::createFromFormat('d/m/Y H:i:s', $date);
            }
            $modulesWithVersion = $this->getDeployedProductVersion($productModules);
            $productId = $deploy->getProduct()->getId();
            $cdCliente = InstanceProvider::getUtilities()->getCurrentCdCliente();
            $productVersion = isset($modulesWithVersion[$productId]) ? $modulesWithVersion[$productId] : null;
            if($productVersion) {
                /** @var ProductVersion $productVersion */
                $productVersion = $this->productVersionRepository->find($productVersion);
                $this->setProductVersionToCustomer($productId, $productVersion);
            }

            $deploy = $this->deployFactory->saveAfterApply($deploy, $isActive, $date, $status, $pathBkp, $productVersion, $ticket);
            $this->entityManager->persist($deploy);

            $this->updateDeployServers($deploy, $status);
            $this->updateChildrenDeploys($deploy, $isActive, $date, $status, $modulesWithVersion, $ticket);

            $this->entityManager->flush();
            $this->updatesSetCdSystemTicket($deployId, $connection, $ticket);
            $this->updateDeployStepServer($deploy, $date);

            try {
                $this->sendCommentsApplyDeploy($server, $version, $deployId, $cdCliente, 'S', $productModules);
            } catch (\Exception $e) {}

            $connection->commit();
            return array('ID' => $deploy->getId());
        } catch (\Exception $e) {
            if($connection->isTransactionActive()) {
                $connection->rollBack();
            }
            throw $e;
        }
    }

    private function sendCommentsApplyDeploy($server, $version, $deployId, $cdCliente, $isClient, $productModules = []){
        $pmIntegration = InstanceProvider::getService()->getPmIntegration();
        $deployRepository = InstanceProvider::getRepository()->getDeploy();
        $deployData = $deployRepository->findDeployUpdates('', $cdCliente, $deployId, $isClient);
        $products = array();
        foreach ($productModules as $product) {
            $productInfo['ISSUES'] = array();

            foreach($deployData as $task){

                $productInfo['VERSION'] = $product['AVAILABLE_VERSION'];

                if($product['ID'] == $task['PRODUCT_ID']){
                    if($task['IS_ISSUE'] == 'N') {
                    $relation = InstanceProvider::getRepository()->getTaskRelationship()->findOneBy(array('taskRelationship' => $task['ID'], 'type' => 'I'));
                        if($relation) {
                            array_push($productInfo['ISSUES'], $relation->getTask()->getId());
                            $products[$product['NAME']] = $productInfo;
                        }
                    } else {
                        array_push($productInfo['ISSUES'], $task['ID']);
                        $products[$product['NAME']] = $productInfo;
                    }
                }
            }
        }

        if(!empty($products)) {
           $pmIntegration->sendProductApplyDeploy($server, $products);

        }
    }

    private function updateDeployStepServer(Deploy $deploy, $date) {
        $deployStepServerRepository = InstanceProvider::getRepository()->getDeployStepServer();
        /** @var \Zeedhi\DevOps\Model\Configurationmanager\Entities\DeployStepServer $deployStepServer */
        $deployStepServer = $deployStepServerRepository->findOneBy(array('deploy' => $deploy));
        if($deployStepServer) {
            $deployStep = $deployStepServer->getDeployStep();
            $pendingDeploys = $deployStepServerRepository->findDeployStepServerStatus($deployStep->getId(),
                array(Deploy::PENDING_DEPLOY_STATUS));
            $errorDeploys   = $deployStepServerRepository->findDeployStepServerStatus($deployStep->getId(),
                array(Deploy::ERROR_DEPLOY_STATUS));
            $appliedDeploys = $deployStepServerRepository->findDeployStepServerStatus($deployStep->getId(),
                array(Deploy::DEPLOYED_DEPLOY_STATUS));

            if(empty($appliedDeploys)) {
                $status = empty($errorDeploys) ? Deploy::PENDING_DEPLOY_STATUS : Deploy::ERROR_DEPLOY_STATUS;
            } else {
                $status = empty($errorDeploys) && empty($pendingDeploys) ? Deploy::DEPLOYED_DEPLOY_STATUS : Deploy::PARTIAL_DEPLOY_STATUS;
            }

            $maxDate = InstanceProvider::getRepository()->getDeployStep()->findMaxDeployStepServerDate($deployStep->getId());
            $maxDate = empty($maxDate) ? null : \DateTime::createFromFormat('d/m/Y H:i:s', $maxDate[0]['DEPLOY_DATE']);
            $date    = $maxDate ? $maxDate : $date;
            if($date) {
                $deployStep->setDeployDate($date);
            }

            $deployStep->setStatus($status);
            $this->entityManager->persist($deployStep);
            $this->entityManager->flush();
        }
    }

    /**
     * @param $productId
     * @param ProductVersion $productVersion
     * @throws \Exception
     */
    private function setProductVersionToCustomer($productId, $productVersion) {
        /** @var ProductVersion $currentRevision */
        $currentRevision = InstanceProvider::getRepository()->getProductVersion()->findOneBy(array('version' => $productVersion->getVersion(),
            'revision' => $productVersion->getRevision(), 'product' => $productId, 'isRevision' => ProductVersion::IS_REVISION_YES));
        $productVersionId = $currentRevision->getId();
        $omProdutoOrgVersaoRepository = InstanceProvider::getRepository()->getOmProdutoOrgVersao();
        $customerVersions = $omProdutoOrgVersaoRepository->findRelatedCustomerVersionsByProduct (
            $productId, $this->utilities->getCurrentCdCliente()
        );
        if(!empty($customerVersions)) {
            $productVersionIds = Utilities::getArrayFromAssociativeArray($customerVersions, 'PRODUCT_VERSION_ID');
            if(in_array($productVersionId, $productVersionIds)) {
                foreach ($customerVersions as $customerVersion) {
                    if($customerVersion['PRODUCT_VERSION_ID'] == $productVersionId && $customerVersion['IDATIVO'] == 'N') {
                        $omProdutoOrgVersao = $omProdutoOrgVersaoRepository->find($customerVersion['ID']);
                        $omProdutoOrgVersao->setIdativo('S');
                        $this->entityManager->persist($omProdutoOrgVersao);
                        break;
                    }
                }
            } else {
                $nrOrg      = $this->utilities->getCurrentOrganizationId();
                $cdOperador = $this->utilities->getCurrentUserId();
                $actualDate = $actualDate = new \DateTime();
                $omProdutoOrgVersao = InstanceProvider::getFactory()->getOmProdutoOrgVersao()->save(null, $customerVersions[0]['OM_PRODUTO_ORG_ID'],
                    $productVersionId, 'S', $actualDate, $cdOperador, $nrOrg);
                $this->entityManager->persist($omProdutoOrgVersao);
            }
        }
    }

    /**
     * @param $productModules
     * @return array
     */
    private function getDeployedProductVersion($productModules) {
        $modulesWithVersion = array();
        foreach($productModules as $module) {
            if($module['IS_UPDATE_VERSION']) {
                $modulesWithVersion[$module['ID']] = $module['PRODUCT_VERSION_ID'];
            } else {
                $version = explode('.', $module['ACTUAL_VERSION']);
                /** @var ProductVersion $productVersion */
                $productVersion = $this->productVersionRepository->findOneBy(array(
                    'version'  => $version[0],
                    'revision' => $version[1],
                    'build'    => $version[2],
                    'product'  => $module['ID']
                ));
                if($productVersion) {
                    $modulesWithVersion[$module['ID']] = $productVersion->getId();
                }
            }
        }
        return $modulesWithVersion;
    }

    /**
     * @param $deploy
     * @param $status
     */
    private function updateDeployServers($deploy, $status) {
        /** @var DeployServer[] $deployServers */
        $deployServers = $this->deployServerRepository->findBy(array('deploy' => $deploy));
        foreach($deployServers as $deployServer) {
            $deployServer->setStatus($status);
            $this->entityManager->persist($deployServer);
        }
    }

    /**
     * @param $deploy
     * @param $isActive
     * @param $date
     * @param $status
     * @param $modulesWithVersion
     * @param $ticket
     */
    private function updateChildrenDeploys($deploy, $isActive, $date, $status, $modulesWithVersion, $ticket) {
        /** @var Deploy[] $childrenDeploys */
        $childrenDeploys = $this->deployRepository->findBy(array('parent' => $deploy));
        foreach($childrenDeploys as $childrenDeploy) {
            $productId = $childrenDeploy->getProduct()->getId();
            $productVersion = isset($modulesWithVersion[$productId]) ? $modulesWithVersion[$productId] : null;
            if($productVersion) {
                $productVersion = $this->productVersionRepository->find($productVersion);
            }
            $childrenDeploy = $this->deployFactory->saveAfterApply($childrenDeploy, $isActive, $date, $status, null, $productVersion, $ticket);
            $this->entityManager->persist($childrenDeploy);

            $this->updatesSetCdSystemTicket($childrenDeploy->getId(), $this->entityManager->getConnection(), $ticket);
        }
    }

    /**
     * @param $repositoryId
     * @param $deployId
     * @param $fileNumber
     * @param $lastFileNumber
     * @return array
     * @throws \Doctrine\ORM\OptimisticLockException
     */
    public function saveInitialMigrationScript($repositoryId, $deployId, $fileNumber, $lastFileNumber) {
        /** @var Deploy $deploy */
        $deploy = InstanceProvider::getRepository()->getDeploy()->find($deployId);
        /** @var \Zeedhi\DevOps\Model\Configurationmanager\Entities\DeployRepository $deployRepository */
        $deployRepository = InstanceProvider::getRepository()->getDeployRepository()->findOneBy(array(
            'deploy' => $deploy, 'externalRepositoryId' => $repositoryId
        ));

        if(!$deployRepository) {
            $fileNumber = $lastFileNumber;
        }

        $deployRepository = InstanceProvider::getFactory()->getDeployRepository()->save($deployRepository, $deploy,
            $repositoryId, $fileNumber);
        $this->entityManager->persist($deployRepository);
        $this->entityManager->flush();

        return array($deployRepository->getId());
    }

    /**
     * @param $productId
     * @return array
     */
    public function findAvailableVersions($productId) {
        $customerCode = $this->utilities->getCurrentCdCliente();
        return InstanceProvider::getRepository()->getOmProdutoOrgVersao()->findAvailableVersionsByProductAndCustomer($productId, $customerCode);
    }

    /**
     * @param $log
     * @return array
     */
    public function findDeployErrors($conditions, $log) {
        $filters = Utilities::createQueryFiltered($conditions);
        $errors = InstanceProvider::getRepository()->getDeploy()->getKnowErrors($filters);
        $knowErrors = [];

        foreach($errors as $error){
            if(strpos($log, $error["DESCRIPTION"])){
                array_push($knowErrors, $error);
            };
        }

        return $knowErrors;
    }

    public function validateTicket($ticket) {
        $return = [[ "COUNT" => 0 ]];
        if(InstanceProvider::getService()->getMovideskIntegration()->validateTicket($ticket)){
            $return = [[ "COUNT" => 1 ]];
        }
        return $return;
    }

    public function findTicketMovDeskDeploy() {
        return InstanceProvider::getService()->getMovideskIntegration()->findTicketMovDeskDeploy();
    }

    /**
     * @param $migrationConfig
     * @return array
     */
    public function getMigrationInfo(array $migrationConfig) {
        foreach ($migrationConfig as &$migration) {
            $migration['MIG_FILES_CONTENT'] = array();
            $migration['MIG_HIGHER_FILE_NUMBER'] = InstanceProvider::getRepository()->getMtDesignEstrutura()->getMaxMigrationNumber($migration['MIG_DESCRIPTION']);
            $pendFiles = empty($migration['MIG_PENDENT_FILES']) ? [0] : $migration['MIG_PENDENT_FILES'];
            $pendFiles = implode(',', $pendFiles);
            $scripts   = InstanceProvider::getRepository()->getMtDesignEstrutura()->getDeployMigration($migration['MIG_DESCRIPTION'],
                $migration['MIG_LAST_FILE_NUMBER'], $pendFiles);
            foreach ($scripts as $scriptRegister) {
                $idx = 0; $notEmptyScripts = array();
                $fullScript = explode("\n", trim($scriptRegister['SCRIPT']));
                foreach ($fullScript as $script) {
                    $trimScript = trim($script);
                    if ($trimScript  !== '') {
                        if ($script[0] == '/' && (!isset($script[1]) || $script[1] != '*')) {
                            $idx++;
                        } else {
                            $notEmptyScripts[$idx] = isset($notEmptyScripts[$idx]) ? $notEmptyScripts[$idx] . "\n" . $script : $script;
                        }
                    }
                }
                $notEmptyScripts = array_values($notEmptyScripts);
                $migration['MIG_FILES_CONTENT'][] = array(
                    'FILE_NAME'     => $scriptRegister['NRSCRIPT'] . '.sql',
                    'FILE_CONTENTS' => $notEmptyScripts,
                    'IS_PEND_FILE'  => $scriptRegister['IDARQPENDENTE'] == 'S'
                );
                $migration['MIG_PENDENT_FILES'] = array_filter($migration['MIG_PENDENT_FILES'], function($pendFile) use ($scriptRegister) {
                    return $pendFile != $scriptRegister['NRSCRIPT'];
                });
                $migration['MIG_PENDENT_FILES'] = array_values($migration['MIG_PENDENT_FILES']);
            }
        }
        return $migrationConfig;
    }

    /**
     * @param $conditions
     * @return array
     */
    public function findDeployUpdates($conditions)
    {
        $deployId = Utilities::getValueInFilterDataSourceByName($conditions, 'DEPLOY_ID', true);
        $isClient = Utilities::getValueInFilterDataSourceByName($conditions, 'IS_CLIENT', true);
        $filterQuery = \Zeedhi\DevOps\Util\Utilities::createQueryFiltered($conditions);
        $cdCliente = $this->utilities->getCurrentCdCliente();
        return InstanceProvider::getRepository()->getDeploy()->findDeployUpdates($filterQuery, $cdCliente, $deployId, $isClient);
    }

    /**
     * @param $deployId
     * @param $conditions
     * @return array
     */
    public function findProductsByDeploy($deployId, $conditions)
    {
        $filterQuery = \Zeedhi\DevOps\Util\Utilities::createQueryFiltered($conditions);
        return InstanceProvider::getRepository()->getDeploy()->findProductsByDeploy($deployId, $filterQuery);
    }

    /**
     * @param $products
     * @param $isClient
     * @return array
     */
    public function findUpdatesBeforeDeploy($products, $isClient)
    {
        $deployRepository = InstanceProvider::getRepository()->getDeploy();
        $query = '';
        $cdcLiente = $this->utilities->getCurrentCdCliente();
        foreach ($products as $product) {
            $query .= $deployRepository->findUpdatesBeforeDeploy($product['id'], $product['version'], $product['revision'], $product['build'],
                                                                 $product['actualVersion'], $product['actualRevision'], $product['actualBuild'],
                                                                 $product['productVersionId'], $product['actualProductVersionId'], $cdcLiente, $isClient);
        }
        $query = substr($query, 0,-5);
        $query = "SELECT * FROM  ({$query}) ORDER BY PV_VERSION, PV_REVISION, PV_BUILD";
        return $deployRepository->applyQueryFindProductsByDeploy($query);
    }


    /**
     * @param $id
     * @param $filter
     * @param $page
     * @param $pageSize
     * @return array
     */
    public function deployStep($id, $filter, $page, $pageSize) {
        if($id) {
            $registers = InstanceProvider::getRepository()->getDeployStep()->findDeployStepById($id);
            foreach ($registers as &$deployStep) {
                $deployStep['SERVER'] = explode(',', $deployStep['SERVER']);
                $serverInfo = InstanceProvider::getRepository()->getDeployStepServer()->findDeployStepServerInfo($deployStep['ID']);
                foreach ($serverInfo as $server) {
                    $id = $server['PRODUCT_SERVER_ID'];
                    $deployStep['SERVER_ID_' .        $id] = $server['SERVER_ID'];
                    $deployStep['SERVER_NAME_' .      $id] = $server['SERVER_NAME'];
                    $deployStep['SERVER_BASE_PATH_' . $id] = $server['PATH'];
                    $deployStep['SERVER_STATUS_' .    $id] = $server['STATUS'];
                    $deployStep['SERVER_IS_ACTIVE_' . $id] = $server['IS_ACTIVE'];
                    $deployStep['SERVER_DATE_' .      $id] = $server['DEPLOY_DATE'];
                    $deployStep['SERVER_DEPLOY_ID_' . $id] = $server['DEPLOY_ID'];
                    $deployStep['SERVER_ORDER_' .     $id] = $server['SERVER_ORDER'];
                    $deployStep['SERVER_FIRST_' .     $id] = $server['FIRST_DEPLOY_STEP'] == 'Y';
                    $deployStep['SERVER_LAST_' .      $id] = $server['LAST_DEPLOY_STEP']  == 'Y';
                    if(!empty($server['TICKET_ID'])) {
                        $deployStep['TICKET_ID']           = $server['TICKET_ID'];
                    }
                }
                if (!empty($deployStep['TICKET_ID'])) {
                    $ticket = InstanceProvider::getService()->getMovideskIntegration()->findTicketById($deployStep['TICKET_ID']);
                    if (!empty($ticket)) {
                        $deployStep['TICKET_NAME'] = $ticket["subject"];
                    }
                }
            }
        } else {
            $limit  = $page * $pageSize;
            $offset = ($page - 1) * $pageSize;
            $customerCode = $this->utilities->getCurrentCdCliente();
            $filterQuery  = \Zeedhi\DevOps\Util\Utilities::createQueryFiltered($filter);
            $registers = InstanceProvider::getRepository()->getDeployStep()->findDeployStep($customerCode, $filterQuery, $limit, $offset);
        }

        return $registers;
    }

    public function saveDeployStep($row) {
        $id   = isset($row['ID']) ? $row['ID'] : null;
        $desc = isset($row['DESCRIPTION']) ? $row['DESCRIPTION'] : null;

        $connection = $this->entityManager->getConnection();
        $this->entityManager->beginTransaction();

        try {
            /** @var DeployStep $deployStep */
            $deployStep = InstanceProvider::getRepository()->getDeployStep()->find($id);
            /** @var \Zeedhi\DevOps\Model\Configurationmanager\Entities\Product $product */
            $product = $this->productRepository->find($row['PRODUCT_ID']);
            /** @var ProductVersion $productVersion */
            $productVersion = $this->productVersionRepository->find($row['PRODUCT_VERSION_ID']);

            if($deployStep) {
                $this->removeNotSelectedServersDeployStep($deployStep, $row['SERVER']);
            }
            $this->validateSaveDeployStep($deployStep, $row['IS_ACTIVE'], $row['SERVER']);

            $deployStep = InstanceProvider::getFactory()->getDeployStep()->save($deployStep, $product, $productVersion,
                $row['STATUS'], $row['IS_ACTIVE'], $desc, $this->utilities->getCurrentCdCliente());
            $this->entityManager->persist($deployStep);
            $this->entityManager->flush();

            $this->saveDeployStepServer($row, $deployStep);
            $this->entityManager->flush();

            $connection->commit();
            return $deployStep->getId();
        } catch(\Exception $e) {
            $connection->rollBack();
            throw $e;
        }
    }

    /**
     * @param $row
     * @param DeployStep $deployStep
     * @throws CustomException
     */
    private function saveDeployStepServer($row, DeployStep $deployStep) {
        $serversOrder = array();
        foreach ($row['SERVER'] as $server) {
            if (in_array($row['SERVER_ORDER_' . $server], $serversOrder)) {
                throw new CustomException('ERR_NOT_POSSIBLE_TO_SAVE_SAME_SERVER_ORDER');
            }
            $serversOrder[] = $row['SERVER_ORDER_' . $server];

            /** @var \Zeedhi\DevOps\Model\Servermanager\Entities\ProductServer $productServer */
            $productServer = $this->productServerRepository->find($server);
            /** @var \Zeedhi\DevOps\Model\Configurationmanager\Entities\DeployStepServer $deployStepServer */
            $deployStepServer = InstanceProvider::getRepository()->getDeployStepServer()->findOneBy(
                array('productServer' => $productServer, 'deployStep' => $deployStep));

            $deployStepServer = InstanceProvider::getFactory()->getDeployStep()->saveServer($deployStepServer,
                $deployStep, $productServer, $row['SERVER_ORDER_' . $server]);
            $this->entityManager->persist($deployStepServer);
        }
    }

    private function removeNotSelectedServersDeployStep(DeployStep $deployStep, $productIds) {
        $deployStepServers = InstanceProvider::getRepository()->getDeployStepServer()->findBy(array('deployStep' => $deployStep));
        /** @var \Zeedhi\DevOps\Model\Configurationmanager\Entities\DeployStepServer[] $deployStepServers */
        foreach ($deployStepServers as $deployStepServer) {
            if(!in_array($deployStepServer->getProductServer()->getId(), $productIds)) {
                $this->entityManager->remove($deployStepServer);
            }
        }
    }

    private function validateSaveDeployStep(DeployStep $deployStep = null, $isActive, $servers) {
        $productServerIds = join(', ', $servers);
        $deploySteps      = InstanceProvider::getRepository()->getDeployStep()->findPendingDeployStepsByProductServer($productServerIds);

        if(!empty($deploySteps)) {
            if(empty($deployStep)) {
                throw new CustomException('ALERT_EXISTS_OTHER_DEPLOY_STEP_WITH_SERVER_CREATE');
            } else if($deployStep->getIsActive() == 'N' && $isActive == 'Y') {
                throw new CustomException('ALERT_EXISTS_OTHER_DEPLOY_STEP_WITH_SERVER_ACTIVATE');
            }
        }
    }

    public function removeDeployStep($id) {
        $connection = $this->entityManager->getConnection();
        $this->entityManager->beginTransaction();
        try {
            /** @var DeployStep $deployStep */
            $deployStep = InstanceProvider::getRepository()->getDeployStep()->find($id);
            /** @var \Zeedhi\DevOps\Model\Configurationmanager\Entities\DeployStepServer[] $deployStepServer */
            $deployStepServer = InstanceProvider::getRepository()->getDeployStepServer()->findBy(array('deployStep' => $deployStep));
            foreach ($deployStepServer as $server) {
                $this->entityManager->remove($server);
                if($server->getDeploy()) {
                    $this->entityManager->remove($server->getDeploy());
                }
            }
            $this->entityManager->flush();

            $this->entityManager->remove($deployStep);
            $this->entityManager->flush();
            $connection->commit();
        } catch(\Exception $e) {
            $connection->rollBack();
            throw $e;
        }
    }

    public function applyDeployStepInfo($deployStepId, $isReapplyDeploy) {
        /** @var DeployStep $deployStep */
        $deployStep     = InstanceProvider::getRepository()->getDeployStep()->find($deployStepId);
        $organizationId = InstanceProvider::getEnvironment()->getCurrentOrganizationId();

        $serverInfo     = InstanceProvider::getRepository()->getDeployStepServer()->findCurrentDeployStepServerInfo($deployStepId, $isReapplyDeploy)[0];
        $productVersion = InstanceProvider::getRepository()->getProductVersion()->findCurrentBuildFromVersion($deployStep->getProduct()->getId(),
            $organizationId, $deployStep->getProductVersion()->getVersion(), $deployStep->getProductVersion()->getRevision())[0];

        $serverInfo['PRODUCT_VERSION_ID'] = $productVersion['ID'];
        $serverInfo['PRODUCT_ID']         = $deployStep->getProduct()->getId();
        $serverInfo['PREVIOUS_DEPLOY']    = [];

        $previousDeploy = $isReapplyDeploy ? null : InstanceProvider::getRepository()->getDeployStepServer()->findPreviousDeployStepServerId($deployStepId);
        if(!empty($previousDeploy)) {
            if(!empty($previousDeploy[0]['MAIN_DEPLOY_ID'])) {
                $previousMainAppliedDeploy = InstanceProvider::getRepository()->getDeploy()->findAppliedDeployInfo($previousDeploy[0]['MAIN_DEPLOY_ID']);
                foreach ($previousMainAppliedDeploy as $mainDeploy) {
                    $serverInfo['PREVIOUS_DEPLOY'][$mainDeploy['PRODUCT_ID']] = $mainDeploy['PRODUCT_VERSION_ID'];
                }    
            }
            $previousAppliedDeploy = InstanceProvider::getRepository()->getDeploy()->findAppliedDeployInfo($previousDeploy[0]['ID']);
            foreach ($previousAppliedDeploy as $deploy) {
                $serverInfo['PREVIOUS_DEPLOY'][$deploy['PRODUCT_ID']] = $deploy['PRODUCT_VERSION_ID'];
            }
        }

        return $serverInfo;
    }

    /**
     * @param $deployStepServerId
     * @param $deployId
     */
    private function setDeployInDeployStepServer($deployStepServerId, $deployId)
    {
        /** @var \Zeedhi\DevOps\Model\Configurationmanager\Entities\DeployStepServer $deployStepServer */
        $deployStepServer = InstanceProvider::getRepository()->getDeployStepServer()->find($deployStepServerId);
        /** @var Deploy $deploy */
        $deploy = $this->deployRepository->find($deployId);

        $deploy->setDeployStepServer($deployStepServer);
        $this->entityManager->persist($deploy);

        $deployStepServer->setDeploy($deploy);
        $this->entityManager->persist($deployStepServer);
    }

    private function setDeployStepPackages(&$deploy) {
        /** @var \Zeedhi\DevOps\Model\Configurationmanager\Entities\DeployPackage[] $deployPackages */
        $deployPackages = $this->deployPackageRepository->findBy(array('deploy' => $deploy['ID']));
        foreach($deployPackages as $deployPackage) {
            $prodVersionPackage = $deployPackage->getProductVersionPackage();
            $deploy['PACKAGE'][] = array(
                'ID' => $prodVersionPackage->getId(),
                'URL' => $prodVersionPackage->getUrl(),
                'NAME' => $prodVersionPackage->getName(),
                'FILE_NAME' => $prodVersionPackage->getFileName(),
                'PATH_TO_DEPLOY' => $prodVersionPackage->getPathToDeploy()
            );
        }
    }

    public function returnBkpDeployStepInfo($deployStepId, $productServerId) {
        $deploy       = $this->deployRepository->findServerInfoReturnBkpDeployStep($deployStepId, $productServerId)[0];
        $mainDeployId = $deploy['MAIN_DEPLOY_ID'];
        $nextDeploy   = null;
        if(!empty($mainDeployId)) {
            $nextDeploy = $deploy;
            $deploy = $this->deployRepository->findServerInfoReturnBkpDeployStepByDeployId($deployStepId, $mainDeployId)[0];
        }
        
        $this->setDeployStepPackages($deploy);
        if($nextDeploy) {
            $this->setDeployStepPackages($nextDeploy);
        }
        
        $deploy['NEXT_DEPLOY'] = $nextDeploy;
        return array($deploy);
    }

    public function updateDeployStepRemoveDeploy($deployId) {
        $connection = $this->entityManager->getConnection();
        $connection->beginTransaction();
        try {
            /** @var Deploy $deploy */
            $deploy = $this->deployRepository->find($deployId);
            $status = Deploy::REMOVED_DEPLOY_STATUS;
            $date   = new \DateTime();

            $deploy->setStatus($status);
            $deploy->setPathBackup(null);
            $deploy->setDeployDate($date);
            $this->entityManager->persist($deploy);

            $this->updateDeployServers($deploy, $status);
            $this->updateDeploysByParent($deploy, $status, $date);

            $this->entityManager->flush();
            $connection->commit();
        } catch (\Exception $e) {
            $connection->rollBack();
            throw $e;
        }
    }

    private function updateDeploysByParent($deploy, $status, $date = null, $pathBackup = null) {
        /** @var Deploy[] $childrenDeploys */
        $childrenDeploys = $this->deployRepository->findBy(array('parent' => $deploy));
        foreach($childrenDeploys as $childrenDeploy) {
            $childrenDeploy->setStatus($status);
            $childrenDeploy->setPathBackup($pathBackup);
            $childrenDeploy->setDeployDate($date);
            $this->entityManager->persist($childrenDeploy);
        }
    }

    public function getNewsDeployStepFilter($deployStepId, $productServerId) {
        $maxVersion = InstanceProvider::getRepository()->getDeployStep()->findMaxDeployStepVersion($deployStepId, $productServerId);
        $minVersion = InstanceProvider::getRepository()->getDeployStep()->findMinDeployStepVersion($deployStepId, $productServerId);
        $filter     = array();

        foreach ($maxVersion as $idx => $max) {
            $filter[$idx] = array(
                'id' => $max['PRODUCT_ID'],

                'version'          => $max['VERSION'],
                'revision'         => $max['REVISION'],
                'build'            => $max['BUILD'],
                'productVersionId' => $max['ID'],

                'actualVersion'          => null,
                'actualRevision'         => null,
                'actualBuild'            => null,
                'actualProductVersionId' => null
            );
            foreach ($minVersion as $min) {
                if($min['PRODUCT_ID'] == $max['PRODUCT_ID']) {
                    $filter[$idx]['actualVersion']          = $min['VERSION'];
                    $filter[$idx]['actualRevision']         = $min['REVISION'];
                    $filter[$idx]['actualBuild']            = $min['BUILD'];
                    $filter[$idx]['actualProductVersionId'] = $min['ID'];
                }
            }
        }

        return $filter;
    }

    public function findModulesConfiguration($productVersionId, $productId)
    {
        $productVersionRepository = InstanceProvider::getRepository()->getProductVersion();
        $productVersion = $productVersionRepository->find($productVersionId);
        $modulesConfiguration = $productVersion->getModulesConfiguration();
        $cdcliente = Utilities::getCdClienteBase();
        $allowedProducts = InstanceProvider::getRepository()->getOmProduto()->allowedProduct($productId, $cdcliente);
        $allowedProducts = Utilities::getArrayFromAssociativeArray($allowedProducts, 'NREXTERNAL_PRODUCT_ID');

        $modules = [];
        if($modulesConfiguration) {
            $modulesConfiguration = json_decode($modulesConfiguration, true);
            foreach($modulesConfiguration as $key => $module) {
                $onlyControl = isset($module['onlyControl']) ? $module['onlyControl'] : false;
                if(!$onlyControl) {
                    if($module['devOpsId'] != $productId AND in_array($module['devOpsId'], $allowedProducts)){
                        $modules[] = $module['devOpsId'];

                    }
                }
            }
        }
        return $modules;
    }

    public function findLatestProductVersion($productId)
    {
        return InstanceProvider::getRepository()->getProductVersion()->findMaxProductVersionWithFilter($productId, '');
    }

    public function getDowngradeVersion($module, $versionKey = 'ACTUAL_VERSION', $avVersionKey = 'AVAILABLE_VERSION') 
    {
        if($module[$versionKey] !== '-' && $module[$avVersionKey] !== '-') {
            $v = explode('.', $module[$versionKey]);
            $av = explode('.', $module[$avVersionKey]);
            $version = intval(str_pad($v[0], 5, '0', STR_PAD_LEFT) . str_pad($v[1], 5, '0', STR_PAD_LEFT) . str_pad($v[2], 5, '0', STR_PAD_LEFT));
            $availableVersion = intval(str_pad($av[0], 5, '0', STR_PAD_LEFT) . str_pad($av[1], 5, '0', STR_PAD_LEFT) . str_pad($av[2], 5, '0', STR_PAD_LEFT));
            return $version > $availableVersion ? 'Y' : 'N';
        }
        return 'N';
    }

    public function getActualModulesProductVersions($modulesVersion) 
    {
        foreach($modulesVersion as &$module) {
            if(empty($module['ACTUAL_PRODUCT_VERSION_ID'])) {
                $module['ACTUAL_PRODUCT_VERSION_ID'] = $this->findActualProductVersionId($module['ID'], $module['ACTUAL_VERSION']);
            }
            $module['DOWNGRADE'] = $this->getDowngradeVersion($module);
        }
        return $modulesVersion;
    }
}
